@@ -1,5 +1,6 @@
 # pylint: disable=too-many-lines, too-many-arguments
 """ Container holding all information about simulation and its components"""
+from __future__ import annotations
 from typing import Dict, Tuple, List, Set, Union
 from math import isclose
 
@@ -14,15 +15,17 @@ from .base import cached_property
 from .validators import assert_unique_names, assert_objects_in_sim_bounds
 from .validators import validate_mode_objects_symmetry
 from .geometry import Box
-from .types import Ax, Shapely, FreqBound, GridSize, Axis, annotate_type
-from .grid import Coords1D, Grid, Coords, GridSpec, UniformGrid
+from .types import Ax, Shapely, FreqBound, Axis, annotate_type, Symmetry
+from .grid.grid import Coords1D, Grid, Coords
+from .grid.grid_spec import GridSpec, UniformGrid
 from .medium import Medium, MediumType, AbstractMedium, PECMedium
-from .boundary import BoundarySpec, Symmetry, BlochBoundary, PECBoundary, PMCBoundary
-from .boundary import PML, StablePML, Absorber
+from .boundary import BoundarySpec, BlochBoundary, PECBoundary, PMCBoundary, Periodic
+from .boundary import PML, StablePML, Absorber, AbsorberSpec
 from .structure import Structure
-from .source import SourceType, PlaneWave, GaussianBeam, AstigmaticGaussianBeam
+from .source import SourceType, PlaneWave, GaussianBeam, AstigmaticGaussianBeam, CustomFieldSource
 from .monitor import MonitorType, Monitor, FreqMonitor
-from .monitor import AbstractFieldMonitor, AbstractNear2FarMonitor
+from .monitor import AbstractFieldMonitor, DiffractionMonitor, AbstractFieldProjectionMonitor
+from .data.dataset import Dataset
 from .viz import add_ax_if_none, equal_aspect
 
 from .viz import MEDIUM_CMAP, PlotParams, plot_params_symmetry
@@ -30,7 +33,7 @@ from .viz import plot_params_structure, plot_params_pml, plot_params_override_st
 from .viz import plot_params_pec, plot_params_pmc, plot_params_bloch
 
 from ..version import __version__
-from ..constants import C_0, MICROMETER, SECOND, inf
+from ..constants import C_0, SECOND, inf
 from ..log import log, Tidy3dKeyError, SetupError, ValidationError
 from ..updater import Updater
 
@@ -107,13 +110,6 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
         units=SECOND,
     )
 
-    grid_size: Union[GridSpec, Tuple[GridSize, GridSize, GridSize]] = pydantic.Field(
-        None,
-        title="Grid Size",
-        description="NOTE: 'grid_size' has been replaced by 'grid_spec'.",
-        units=MICROMETER,
-    )
-
     medium: MediumType = pydantic.Field(
         Medium(),
         title="Background Medium",
@@ -147,9 +143,11 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
     )
 
     boundary_spec: BoundarySpec = pydantic.Field(
-        BoundarySpec(),
+        None,
         title="Boundaries",
-        description="Specification of boundary conditions along each dimension.",
+        description="Specification of boundary conditions along each dimension. If ``None``, "
+        "periodic boundary conditions are applied on all sides. Default will change to PML in 2.0 "
+        "so explicitly setting the boundaries is recommended.",
     )
 
     monitors: Tuple[annotate_type(MonitorType), ...] = pydantic.Field(
@@ -190,11 +188,12 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
     )
 
     courant: float = pydantic.Field(
-        0.9,
+        0.99,
         title="Courant Factor",
         description="Courant stability factor, controls time step to spatial step ratio. "
         "Lower values lead to more stable simulations for dispersive materials, "
-        "but result in longer simulation times.",
+        "but result in longer simulation times. This factor is normalized to no larger than 1 "
+        "when CFL stability condition is met in 3D.",
         gt=0.0,
         le=1.0,
     )
@@ -207,17 +206,33 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
 
     """ Validating setup """
 
-    @pydantic.validator("grid_size", always=True)
-    def _error_use_grid_size(cls, val):
-        """If ``grid_size`` is provided, raise an error."""
+    @pydantic.root_validator(pre=True)
+    def _update_simulation(cls, values):
+        """Update the simulation if it is an earlier version."""
 
-        if val is not None:
-            raise ValidationError(
-                "'grid_size' has been replaced by 'grid_spec'. See the "
-                "'GridSpec' documentation for more information."
+        # if no version, assume it's already updated
+        if "version" not in values:
+            return values
+
+        # otherwise, call the updator to update the values dictionary
+        updater = Updater(sim_dict=values)
+        return updater.update_to_current()
+
+    # TODO: remove for 2.0
+    @pydantic.root_validator(pre=True)
+    def _deprecation_2_0_missing_defaults(cls, values):
+        """Raise deprecation warning if a default ``boundary_spec`` is used."""
+
+        if values.get("boundary_spec") is None:
+            log.warning(
+                "'Simulation.boundary_spec' uses default value, which is 'Periodic()' on all "
+                "sides but will change to 'PML()' in Tidy3D version 2.0. "
+                "We recommend explicitly setting all boundary conditions "
+                "ahead of this release to avoid unexpected results."
             )
+            values["boundary_spec"] = BoundarySpec.all_sides(boundary=Periodic())
 
-        return val
+        return values
 
     @pydantic.validator("grid_spec", always=True)
     def _validate_auto_grid_wavelength(cls, val, values):
@@ -232,17 +247,11 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
     _mode_sources_symmetries = validate_mode_objects_symmetry("sources")
     _mode_monitors_symmetries = validate_mode_objects_symmetry("monitors")
 
-    # assign names to unnamed structures, sources, and mediums
-    # _structure_names = set_names("structures")
-    # _source_names = set_names("sources")
-
     # make sure all names are unique
     _unique_structure_names = assert_unique_names("structures")
     _unique_source_names = assert_unique_names("sources")
     _unique_monitor_names = assert_unique_names("monitors")
 
-    # _unique_medium_names = assert_unique_names("structures", check_mediums=True)
-
     # _few_enough_mediums = validate_num_mediums()
     # _structures_not_at_edges = validate_structure_bounds_not_at_edges()
     # _gap_size_ok = validate_pml_gap_size()
@@ -263,6 +272,70 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
                 )
         return val
 
+    # pylint: disable=too-many-locals
+    @pydantic.validator("boundary_spec", always=True)
+    def plane_wave_boundaries(cls, val, values):
+        """Error if there are plane wave sources incompatible with boundary conditions."""
+        boundaries = val.to_list
+        sources = values.get("sources")
+        size = values.get("size")
+        sim_medium = values.get("medium")
+        structures = values.get("structures")
+        for src_idx, source in enumerate(sources):
+            if not isinstance(source, PlaneWave):
+                continue
+
+            _, tan_dirs = cls.pop_axis([0, 1, 2], axis=source.injection_axis)
+            medium_set = cls.intersecting_media(source, structures)
+            medium = medium_set.pop() if medium_set else sim_medium
+
+            for tan_dir in tan_dirs:
+                boundary = boundaries[tan_dir]
+
+                # check the PML/absorber + angled plane wave case
+                num_pml = sum(isinstance(bnd, AbsorberSpec) for bnd in boundary)
+                if num_pml > 0 and source.angle_theta != 0:
+                    raise SetupError(
+                        "Angled plane wave sources are not compatible with the absorbing boundary "
+                        f"along dimension {tan_dir}. Either set the source ``angle_theta`` to "
+                        "``0``, or use Bloch boundaries that match the source angle."
+                    )
+
+                # check the Bloch boundary + angled plane wave case
+                num_bloch = sum(isinstance(bnd, BlochBoundary) for bnd in boundary)
+                if num_bloch > 0:
+                    # make a dummy Bloch boundary to check for correctness
+                    dummy_bnd = BlochBoundary.from_source(
+                        source=source, domain_size=size[tan_dir], axis=tan_dir, medium=medium
+                    )
+                    expected_bloch_vec = dummy_bnd.bloch_vec
+
+                    if boundary[0].bloch_vec != expected_bloch_vec:
+                        test_val = np.abs(expected_bloch_vec - boundary[0].bloch_vec)
+
+                        if np.isclose(test_val % 1, 0) and not np.isclose(test_val, 0):
+                            # the given Bloch vector is offset by an integer
+                            log.warning(
+                                f"The wave vector of source at index {src_idx} along dimension "
+                                f"{tan_dir} is equal to the Bloch vector of the simulation "
+                                "boundaries along that dimension plus an integer reciprocal "
+                                "lattice vector. If using a ``DiffractionMonitor``, diffraction "
+                                "order 0 will not correspond to the angle of propagation "
+                                "of the source. Consider using ``BlochBoundary.from_source()``."
+                            )
+                        elif not np.isclose(test_val % 1, 0):
+                            # the given Bloch vector is neither equal to the expected value, nor
+                            # off by an integer
+                            log.warning(
+                                f"The Bloch vector along dimension {tan_dir} may be incorrectly "
+                                f"set with respect to the source at index {src_idx}. The absolute "
+                                "difference between the expected and provided values in "
+                                "bandstructure units, up to an integer offset, is greater than "
+                                "1e-6. Consider using ``BlochBoundary.from_source()``, or "
+                                "double-check that it was defined correctly."
+                            )
+        return val
+
     @pydantic.validator("structures", always=True)
     def _validate_num_mediums(cls, val):
         """Error if too many mediums present."""
@@ -438,18 +511,103 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
         return val
 
     @pydantic.validator("monitors", always=True)
-    def n2f_monitors_with_symmetry(cls, val, values):
-        """Error if a server-side near-to-far monitor is used with symmetry"""
+    def diffraction_monitor_boundaries(cls, val, values):
+        """If any :class:`DiffractionMonitor` exists, ensure boundary conditions in the
+        transverse directions are periodic or Bloch."""
         monitors = val
-        symmetry = values.get("symmetry")
-        if not any(symmetry):
-            return val
+        boundary_spec = values.get("boundary_spec")
         for monitor in monitors:
-            if isinstance(monitor, AbstractNear2FarMonitor):
-                raise SetupError(
-                    "Server-side near-to-far calculation is not available with symmetry. "
-                    "Use the local near-to-far available through the 'RadiationVectors' class."
-                )
+            if isinstance(monitor, DiffractionMonitor):
+                _, (n_x, n_y) = monitor.pop_axis(["x", "y", "z"], axis=monitor.normal_axis)
+                boundaries = [
+                    boundary_spec[n_x].plus,
+                    boundary_spec[n_x].minus,
+                    boundary_spec[n_y].plus,
+                    boundary_spec[n_y].minus,
+                ]
+                # make sure the transverse boundaries are either periodic or Bloch
+                for boundary in boundaries:
+                    if not isinstance(boundary, (Periodic, BlochBoundary)):
+                        raise SetupError(
+                            f"The 'DiffractionMonitor' {monitor.name} requires periodic "
+                            f"or Bloch boundaries along dimensions {n_x} and {n_y}."
+                        )
+        return val
+
+    # pylint: disable=too-many-locals
+    @pydantic.validator("monitors", always=True)
+    def _projection_monitors_homogeneous(cls, val, values):
+        """Error if any field projection monitor is not in a homogeneous region."""
+
+        if val is None:
+            return val
+
+        # list of structures including background as a Box()
+        structure_bg = Structure(
+            geometry=Box(
+                size=values.get("size"),
+                center=values.get("center"),
+            ),
+            medium=values.get("medium"),
+        )
+
+        structures = values.get("structures") or []
+        total_structures = [structure_bg] + list(structures)
+
+        for monitor in val:
+            if isinstance(monitor, (AbstractFieldProjectionMonitor, DiffractionMonitor)):
+                mediums = cls.intersecting_media(monitor, total_structures)
+                # make sure there is no more than one medium in the returned list
+                if len(mediums) > 1:
+                    raise SetupError(
+                        f"{len(mediums)} different mediums detected on plane "
+                        f"intersecting a {monitor.type}. Plane must be homogeneous."
+                    )
+
+        return val
+
+    @pydantic.validator("monitors", always=True)
+    def _projection_monitors_distance(cls, val, values):
+        """Warn if the projection distance is large for exact projections."""
+
+        if val is None:
+            return val
+
+        sim_size = values.get("size")
+
+        for idx, monitor in enumerate(val):
+            if isinstance(monitor, AbstractFieldProjectionMonitor):
+                if (
+                    np.abs(monitor.proj_distance) > 1.0e4 * np.max(sim_size)
+                    and not monitor.far_field_approx
+                ):
+                    monitor = monitor.copy(update={"far_field_approx": True})
+                    val = list(val)
+                    val[idx] = monitor
+                    val = tuple(val)
+                    log.warning(
+                        "A very large projection distance was set for the field projection "
+                        f"monitor '{monitor.name}'. Using exact field projections may result in "
+                        "precision loss for large distances; automatically enabling far-field "
+                        "approximations ('far_field_approx = True') for better precision. "
+                        "To insist on exact projections, consider using client-side projections "
+                        "via the 'FieldProjector' class, where higher precision is available."
+                    )
+        return val
+
+    @pydantic.validator("monitors", always=True)
+    def diffraction_monitor_medium(cls, val, values):
+        """If any :class:`DiffractionMonitor` exists, ensure is does not lie in a lossy medium."""
+        monitors = val
+        structures = values.get("structures")
+        medium = values.get("medium")
+        for monitor in monitors:
+            if isinstance(monitor, DiffractionMonitor):
+                medium_set = Simulation.intersecting_media(monitor, structures)
+                medium = medium_set.pop() if medium_set else medium
+                _, index_k = medium.nk_model(frequency=np.array(monitor.freqs))
+                if not np.all(index_k == 0):
+                    raise SetupError("Diffraction monitors must not lie in a lossy medium.")
         return val
 
     @pydantic.validator("grid_spec", always=True)
@@ -517,18 +675,9 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
         # for each plane wave in the sources list
         for source in val:
             if isinstance(source, (PlaneWave, GaussianBeam, AstigmaticGaussianBeam)):
-
-                # get all merged structures on the plane
-                normal_axis_index = source.size.index(0.0)
-                dim = "xyz"[normal_axis_index]
-                pos = source.center[normal_axis_index]
-                xyz_kwargs = {dim: pos}
-                structures_merged = cls._filter_structures_plane(total_structures, **xyz_kwargs)
-
+                mediums = cls.intersecting_media(source, total_structures)
                 # make sure there is no more than one medium in the returned list
-                mediums = {medium for medium, _ in structures_merged}
                 if len(mediums) > 1:
-
                     raise SetupError(
                         f"{len(mediums)} different mediums detected on plane "
                         f"intersecting a {source.type} source. Plane must be homogeneous."
@@ -558,6 +707,7 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
         """Validate the fully initialized simulation is ok for upload to our servers."""
         self._validate_size()
         self._validate_monitor_size()
+        self._validate_datasets_not_none()
         # self._validate_run_time()
 
     def _validate_size(self) -> None:
@@ -608,13 +758,13 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
                 f"a maximum of {MAX_MONITOR_DATA_SIZE_BYTES:.2e} are allowed."
             )
 
-    def _validate_run_time(self) -> None:
-        """Ensures that the simulation run time is > 0."""
-
-        if self.run_time <= 0:
+    def _validate_datasets_not_none(self) -> None:
+        """Ensures that all custom datasets are defined."""
+        if any(dataset is None for dataset in self.custom_datasets):
             raise SetupError(
-                "The `Simulation.run_time` parameter was left at its default value of 0.0. "
-                "For running a simulation on our servers it must be set to > 0.0."
+                "Data for a custom data component is missing. This can happen for example if the "
+                "Simulation has been loaded from json. To save and load simulations with custom "
+                "data, use hdf5 format instead."
             )
 
     """ Accounting """
@@ -658,6 +808,67 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
         geometry = Box(size=(inf, inf, inf))
         return Structure(geometry=geometry, medium=self.medium)
 
+    @staticmethod
+    def intersecting_media(
+        test_object: Box, structures: Tuple[Structure, ...]
+    ) -> Tuple[MediumType, ...]:
+        """From a given list of structures, returns a list of :class:`AbstractMedium` associated
+        with those structures that intersect with the ``test_object``, if it is a surface, or its
+        surfaces, if it is a volume.
+
+        Parameters
+        -------
+        test_object : :class:`Box`
+            Object for which intersecting media are to be detected.
+        structures : List[:class:`AbstractMedium`]
+            List of structures whose media will be tested.
+
+        Returns
+        -------
+        List[:class:`AbstractMedium`]
+            Set of distinct mediums that intersect with the given planar object.
+        """
+        if test_object.size.count(0.0) == 1:
+            # get all merged structures on the test_object, which is already planar
+            normal_axis_index = test_object.size.index(0.0)
+            dim = "xyz"[normal_axis_index]
+            pos = test_object.center[normal_axis_index]
+            xyz_kwargs = {dim: pos}
+            structures_merged = Simulation._filter_structures_plane(structures, **xyz_kwargs)
+            mediums = {medium for medium, _ in structures_merged}
+            return mediums
+
+        # if the test object is a volume, test each surface recursively
+        object_dict = test_object.dict()
+        exclude_surfaces = object_dict.pop("exclude_surfaces", None)
+        surfaces = test_object.surfaces(**object_dict)
+        if exclude_surfaces:
+            surfaces = [surf for surf in surfaces if surf.name[-2:] not in exclude_surfaces]
+        mediums = set()
+        for surface in surfaces:
+            _mediums = Simulation.intersecting_media(surface, structures)
+            mediums.update(_mediums)
+        return mediums
+
+    def monitor_medium(self, monitor: MonitorType):
+        """Return the medium in which the given monitor resides.
+
+        Parameters
+        -------
+        monitor : :class:`Monitor`
+            Monitor whose associated medium is to be returned.
+
+        Returns
+        -------
+        :class:`AbstractMedium`
+            Medium associated with the given :class:`Monitor`.
+        """
+        medium_set = self.intersecting_media(monitor, self.structures)
+        if len(medium_set) > 1:
+            raise SetupError(f"Monitor {monitor.name} intersects more than one medium.")
+        medium = medium_set.pop() if medium_set else self.medium
+        return medium
+
     """ Plotting """
 
     @equal_aspect
@@ -1308,8 +1519,6 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
             The supplied or created matplotlib axes.
         """
 
-        kwargs.setdefault("clip_on", False)
-
         def set_plot_params(boundary_edge, lim, side, thickness):
             """Return the line plot properties such as color and opacity based on the boundary"""
             if isinstance(boundary_edge, PECBoundary):
@@ -1421,8 +1630,10 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
         _, (xmin, ymin) = self.pop_axis(self.bounds_pml[0], axis=axis)
         _, (xmax, ymax) = self.pop_axis(self.bounds_pml[1], axis=axis)
 
-        ax.set_xlim(xmin, xmax)
-        ax.set_ylim(ymin, ymax)
+        if xmin != xmax:
+            ax.set_xlim(xmin, xmax)
+        if ymin != ymax:
+            ax.set_ylim(ymin, ymax)
         return ax
 
     @staticmethod
@@ -1625,8 +1836,8 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
             Minimum wavelength in the material (microns).
         """
         freq_max = max(source.source_time.freq0 for source in self.sources)
-        wvl_min = C_0 / min(freq_max)
-        eps_max = max(abs(structure.medium.get_eps(freq_max)) for structure in self.structures)
+        wvl_min = C_0 / freq_max
+        eps_max = max(abs(structure.medium.eps_model(freq_max)) for structure in self.structures)
         n_max, _ = AbstractMedium.eps_complex_to_nk(eps_max)
         return wvl_min / n_max
 
@@ -1693,28 +1904,29 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
                 # bmin <= center <= bmax
                 bmin_tmp = center
                 bmax_tmp = max(bmax, 2 * center - bmin)
-            # Extend well past the simulation domain if needed, but truncate if original box
-            # is too large, specifically to avoid issues with inf.
-            sim_size = sim_bmax - sim_bmin
-            bmin_new.append(max(bmin_tmp, sim_bmin - sim_size))
-            bmax_new.append(min(bmax_tmp, sim_bmax + sim_size))
+            # If inf, extend well past the simulation domain but make a finite number
+            if np.isinf(bmin_tmp):
+                bmin_tmp = sim_bmin - np.amax(self.size)
+            if np.isinf(bmax_tmp):
+                bmax_tmp = sim_bmax + np.amax(self.size)
+            bmin_new.append(bmin_tmp)
+            bmax_new.append(bmax_tmp)
 
         return Box.from_bounds(bmin_new, bmax_new)
 
-    def discretize(self, box: Box, extend: bool = False, snap_zero_dim: bool = False) -> Grid:
+    def discretize(self, box: Box, snap_zero_dim: bool = False, **kwargs) -> Grid:
         """Grid containing only cells that intersect with a :class:`Box`.
 
         Parameters
         ----------
         box : :class:`Box`
             Rectangular geometry within simulation to discretize.
-        extend : bool
-            If ``True``, extra pixels are added to the discretized grid if needed, such that Yee
-            grid fields can be interpolated to any point inside the ``box``.
         snap_zero_dim : bool
             If ``True``, and the ``box`` has size zero along a given direction, the ``grid`` is
             defined to also have a zero-sized cell exactly centered at the ``box`` center. If
             false, the ``simulation`` grid cell containing the ``box`` center is instead used.
+        kwargs :
+            Extra keyword arguments passed to ``discretize_inds`` method of :class:`.Grid`.
 
         Returns
         -------
@@ -1725,16 +1937,19 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
         if not self.intersects(box):
             log.error(f"Box {box} is outside simulation, cannot discretize")
 
-        span_inds = self.grid.discretize_inds(box, extend=extend)
+        span_inds = self.grid.discretize_inds(box, **kwargs)
         boundary_dict = {}
         for idim, dim in enumerate("xyz"):
-            if snap_zero_dim and box.size[idim] == 0:
-                boundary_dict[dim] = [box.center[idim], box.center[idim]]
-            else:
-                ind_beg, ind_end = span_inds[idim]
-                # ind_end + 1 because we are selecting cell boundaries not cells
-                boundary_dict[dim] = self.grid.periodic_subspace(idim, ind_beg, ind_end + 1)
+            ind_beg, ind_end = span_inds[idim]
+            # ind_end + 1 because we are selecting cell boundaries not cells
+            boundary_dict[dim] = self.grid.periodic_subspace(idim, ind_beg, ind_end + 1)
 
+            # Overwrite with zero dimension snapped, if requested
+            if snap_zero_dim:
+                if self.size[idim] == 0:
+                    boundary_dict[dim] = [self.center[idim], self.center[idim]]
+                elif box.size[idim] == 0:
+                    boundary_dict[dim] = [box.center[idim], box.center[idim]]
         return Grid(boundaries=Coords(**boundary_dict))
 
     def epsilon(
@@ -1800,14 +2015,14 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
             refer to `xarray's Documentaton <https://tinyurl.com/2zrzsp7b>`_.
         """
 
-        def get_eps(medium: Medium, freq: float):
+        def get_eps(structure: Structure, frequency: float):
             """Select the correct epsilon component if field locations are requested."""
             if coord_key[0] != "E":
-                return medium.eps_model(freq)
+                return self.medium.eps_model(frequency)
             component = ["x", "y", "z"].index(coord_key[1])
-            return medium.eps_diagonal(freq)[component]
+            return structure.eps_diagonal(frequency)[component]
 
-        eps_background = get_eps(self.medium, freq)
+        eps_background = get_eps(structure=self.background_structure, frequency=freq)
 
         def make_eps_data(coords: Coords):
             """returns epsilon data on grid of points defined by coords"""
@@ -1820,7 +2035,7 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
             for structure in self.structures:
                 if not points_box.intersects(structure.geometry):
                     continue
-                eps_structure = get_eps(structure.medium, freq)
+                eps_structure = get_eps(structure=structure, frequency=freq)
                 is_inside = structure.geometry.inside(x, y, z)
                 eps_array[np.where(is_inside)] = eps_structure
             coords = {"x": np.array(xs), "y": np.array(ys), "z": np.array(zs)}
@@ -1830,27 +2045,10 @@ class Simulation(Box):  # pylint:disable=too-many-public-methods
         coords = grid[coord_key]
         return make_eps_data(coords)
 
-    @classmethod
-    def from_file(cls, fname: str, **parse_kwargs):
-        """Loads a :class:`Tidy3dBaseModel` from .yaml or .json file.
-
-        Parameters
-        ----------
-        fname : str
-            Full path to the .yaml or .json file to load the :class:`Tidy3dBaseModel` from.
-        **parse_kwargs
-            Keyword arguments passed to either pydantic's ``parse_file`` or ``parse_raw`` methods
-            for ``.json`` and ``.yaml`` file formats, respectively.
-        Returns
-        -------
-        :class:`Tidy3dBaseModel`
-            An instance of the component class calling `load`.
-
-        Example
-        -------
-        >>> simulation = Simulation.from_file(fname='folder/sim.json') # doctest: +SKIP
+    @property
+    def custom_datasets(self) -> List[Dataset]:
+        """List of custom datasets for verification purposes. If the list is not empty, then
+        the simulation needs to be exported to hdf5 to store the data.
         """
-
-        updater = Updater.from_file(fname)
-        sim_dict = updater.update_to_current()
-        return Simulation.parse_obj(sim_dict, **parse_kwargs)
+        datasets = [src.field_dataset for src in self.sources if isinstance(src, CustomFieldSource)]
+        return datasets
